#include <iostream.h>
#include "bootstrap.hh"
#include "IRread.hh"
#include "ec_dii.hh"
#include "compatibility.hh"


int main( int argc, char **argv) {
  Bootstrap::init(argc,argv);

  IRMOD Repository_var rep;
  CosNaming::NamingContext_var rootns= Bootstrap::getNS();
  try {
    rep= IRMOD Repository::_narrow( rootns->resolve( Bootstrap::SRV_IR()));
  }
  catch (...) {
  }
  if (CORBA::is_nil(rep)) {
    cerr << "Could not find Interface Repository service" << endl;
    return 1;
  }
  IRMOD Contained_var cntd;
  const char *intf_id= "IDL:/TestIntf:1.0";
  try {
    cntd= rep->lookup("::TestIntf");
    if (CORBA::is_nil(cntd)) {
      cerr << "Not found" << endl;
      throw 1;
    }
  }
  catch (CORBA::SystemException &se) {
    cerr << "SystemException" << se << endl;
  }
  catch (CORBA::UserException &ue) {
    cerr << "UserException" << endl;
  }
  catch (...) {
    cerr << "Lookup for " << intf_id << " failed" << endl;
    return 1;
  }

  //  cout << (const char *) Bootstrap::orb_ptr->object_to_string(cntd) << endl;

  IRMOD InterfaceDef_ptr idef;
  try {
    idef= IRMOD InterfaceDef::_narrow(cntd);
    if (CORBA::is_nil(idef)) throw 1;
  }
  catch (CORBA::Exception &ex) {
    cerr << ex << endl;
    return 0;
  }
  catch (...) {
    cerr << intf_id << " is not an interface" << endl;
    return 0;
  }

  cerr << "check_interface" << endl;
  ECHelper::TypedInterface tpintf;
  try {
    if (!check_interface( idef, tpintf)) {
      cerr << "Interface " << intf_id << " does not fulfil contraints for typed event channel" << endl;
    }
  }
  catch (CORBA::SystemException &se) {
    cerr << se << endl;
  }

  CORBA::Object_var remote_obj= 
    Bootstrap::orb_ptr->string_to_object( argv[1]);
  cerr << "about to create dii" << endl;
  EC_DII *ec_dii= default_ec_dii_factory->create_ec_dii( remote_obj,
							 tpintf);
  cerr << "dii created" << endl;
  ECHelper::SeqAny params;
  /*
  params.length(2);
  params[1] <<= (CORBA::Double) 3.14159;
  CORBA::Any::from_boolean fb( 1);
  params[0] <<= fb;
  */
  CORBA::Any a;
  a <<= (CORBA::Double) 2.123123;
  params.length(1);
#ifdef OMNIORB
  CORBA::Any::from_string fs("aldkfjalsdf",0);
  params[0] <<= fs;
#elif defined(ORBIX)
  params[0] <<= (const char *) "alsdfjalsdkf";
#endif
  
  try {
    cerr << "Sending request" << endl;
    ec_dii->send("operString", params);
    cerr << "Request sent" << endl;
  }
  catch (CORBA::Exception &ex) {
    cerr << ex << endl;
  }
  try {
    remote_obj->_non_existent();
  }
  catch (...) {
  }
  return 0;
}
