#include <iostream.h>
#include "IRread.hh"
#include "ec_dsi.hh"
#include "bootstrap.hh"
#include "compatibility.hh"


class My_Dispatcher : public EC_Dispatcher {
public:
  virtual CORBA::Boolean ec_dispatch( CORBA::Object_ptr obj, 
				      CORBA::Boolean twoway,
				      const char *operation_name,
				      ECHelper::SeqAny &params) {
    cerr << "Dispatch operation " << operation_name << endl;
    cerr << "Parameters : " << (int) params.length() << endl;
    for (CORBA::ULong i=0; i<params.length(); i++) {
      cerr << "# " << (int) i << endl;
      cerr << "\tkind " << (int) params[i].type()->kind() << endl;
      if (params[i].type()->kind() == CORBA::tk_double) {
	CORBA::Double d;
	params[i] >>= d;
	cerr << (double) d << endl;
      }
      else if (params[i].type()->kind() == CORBA::tk_string) {
	char *s;
	params[i] >>= s;
	cerr << s << endl;
      }
    } 
    return (CORBA::Boolean) 0;
  }
};

int main( int argc, char **argv) {
#ifdef ORBIX
  CORBA::Orbix.setDiagnostics(0);
#endif
  Bootstrap::init(argc,argv);

  
  IRMOD Repository_var rep;
  CosNaming::NamingContext_var rootns= Bootstrap::getNS();
  try {
    rep= IRMOD Repository::_narrow( rootns->resolve( Bootstrap::SRV_IR()));
  }
  catch (...) {
  }
  if (CORBA::is_nil(rep)) {
    cerr << "Could not find Interface Repository service" << endl;
    return 1;
  }
  IRMOD Contained_var cntd;
  const char *intf_id= "IDL:/TestIntf:1.0";
  try {
    cntd= rep->lookup("::TestIntf");
    if (CORBA::is_nil(cntd)) {
      cerr << "Not found" << endl;
      throw 1;
    }
  }
  catch (CORBA::SystemException &se) {
    cerr << "SystemException" << se << endl;
  }
  catch (CORBA::UserException &ue) {
    cerr << "UserException" << endl;
  }
  catch (...) {
    cerr << "Lookup for " << intf_id << " failed" << endl;
    return 1;
  }

  //  cout << (const char *) Bootstrap::orb_ptr->object_to_string(cntd) << endl;

  IRMOD InterfaceDef_ptr idef;
  try {
    idef= IRMOD InterfaceDef::_narrow(cntd);
    if (CORBA::is_nil(idef)) throw 1;
  }
  catch (CORBA::Exception &ex) {
    cerr << ex << endl;
    return 0;
  }
  catch (...) {
    cerr << intf_id << " is not an interface" << endl;
    return 0;
  }
  
  cerr << "check_interface" << endl;
  ECHelper::TypedInterface tpintf;
  
  try {
       if (!check_interface( idef, tpintf)) {
      cerr << "Interface " << intf_id << " does not fulfil contraints for typed event channel" << endl;
    }
  }
  catch (CORBA::SystemException &se) {
    cerr << se << endl;
  }
  
  My_Dispatcher mdisp;
  EC_DSI *ec_dsi= default_ec_dsi_factory->create_ec_dsi( tpintf, 0, mdisp);
  CORBA::Object_var ecobj= ec_dsi->get_object();
  cout << (const char *) Bootstrap::orb_ptr->object_to_string( ecobj) << endl;
  cerr << "Implementation is ready for service" << endl;
  try {
    IMPL_IS_READY( Bootstrap::boa_ptr);
    //    Bootstrap::boa_ptr->impl_is_ready("DSI");
  }
  catch (CORBA::Exception &ex) {
    cerr << "Exception " << ex << endl;
  }
  return 0;
}
