// $Id: master_forwarder.cc,v 1.2 1999/05/04 19:29:40 mencnar Exp $
//
// $Source: /home/iris/mencnar/ec/CVS/src/cc/master_forwarder.cc,v $
//
// $Log: master_forwarder.cc,v $
// Revision 1.2  1999/05/04 19:29:40  mencnar
// Added a method for synchronizing clocks.
//
// Revision 1.1  1999/02/24 13:01:58  mencnar
// *** empty log message ***
//
//

#ifdef OMNIORB

#include "master_forwarder.hh"
#include "IORlib.hh"
#ifndef __WIN32__
#include <netdb.h>
#endif
#include "trace.hh"
#include "mtime.hh"

#ifdef __sunos__
extern "C" struct hostent *gethostbyname(const char *name);
extern "C" int gethostname(char *name, int namelen);
#endif

//-------------------------------------------------------------Forwarder_i--
void Forwarder_i::
open_event_channel( EventChannel_i &channel,
		    AbstractForwarderInstance inst,
		    CORBA::ULong _channel_id,
		    CORBA::ULong _max_queue_size,
		    ECAdmin::OverflowAction _oa,
		    CORBA::ULong _keep_alive_time) {
  ec_i= &channel;
  channel_id= _channel_id;
  max_queue_size= _max_queue_size;
  overflow_action= _oa;
  keep_alive_time= _keep_alive_time;
  // this operation is called only once before any other operation
  master.insert_forwarder( channel_id, inst);
}

void Forwarder_i::
close_event_channel() {
  master.remove_forwarder( channel_id);
}

int Forwarder_i::
try_forward_consumer( CosEventChannelAdmin::ProxyPushSupplier_ptr push_supp,
		      CORBA::ULong consumer_id,
		      CosEventComm::PushConsumer_ptr push_cons) {
  trace.start(500) << "try_forward_consumer" << '\n';
  ECMcast::Slave_var slave= master.get_slave( push_cons);
  omni_mutex_lock oml(guard);
  if (CORBA::is_nil(slave)) return 0;

  try {
    int ret= (int) slave->forward_push_consumer( unique_description,
						 channel_id,
						 consumer_id,
						 push_supp,
						 push_cons,
						 max_queue_size,
						 overflow_action,
						 keep_alive_time);
    if (ret) forwards[consumer_id]= ForwardedProxy( slave,
						    push_supp);
    return ret;
  }
  catch (...) {
    master.signal_failed_slave( slave);
  }
  return 0;
}

int Forwarder_i::
try_forward_typed_consumer( CosEventChannelAdmin::ProxyPushSupplier_ptr push_supp,
			    CORBA::ULong consumer_id,
			    IRMOD InterfaceDef_ptr idef,
			    CosEventComm::PushConsumer_ptr push_cons) {
  trace.start(500) << "try_forward_typed_consumer" << '\n';
  ECMcast::Slave_var slave= master.get_slave( push_cons);
  omni_mutex_lock oml(guard);
  if (CORBA::is_nil(slave)) return 0;
  try {
    int ret= (int) slave->forward_typed_push_consumer( unique_description,
						       channel_id,
						       consumer_id,
						       idef,
						       push_supp,
						       push_cons,
						       max_queue_size,
						       overflow_action,
						       keep_alive_time);
    if (ret) forwards[consumer_id]= ForwardedProxy( slave,
						    push_supp);
    return ret;
  }
  catch (...) {
    master.signal_failed_slave( slave);
  }
  return 0;
}

void Forwarder_i::
remove_forwarded_consumer( CORBA::ULong consumer_id) {
  omni_mutex_lock oml(guard);
  Forwards::iterator it=
    forwards.find( consumer_id);
  if (it != forwards.end()) {
    try {
      if ((*it).second.valid)
	(*it).second.slave->close_proxy( channel_id, (*it).first, 1);
    }
    catch (...) {
    }
    forwards.erase( it);
  }
}

void Forwarder_i::
slave_failed( ECMcast::Slave_ptr a_slave) {
  Forwards copy_forwards;
  {
    omni_mutex_lock oml(guard);
    copy_forwards= forwards;
    for (Forwards::iterator it= forwards.begin();
	 it != forwards.end();
	 it++) {
      
      if ( (*it).second.slave == a_slave)
	(*it).second.valid= 0;
    }
  }

  for (Forwards::iterator it= forwards.begin();
       it != forwards.end();
       it++) {

    if ( (*it).second.slave == a_slave) {
      (*it).second.proxy->disconnect_push_supplier();
    }
  }
}

void Forwarder_i::
push( ECDHandler &ecdh) {
  GeneralLock::auto_write gawl(* msource.guard);
  ECMcast::Data data;
  data.channel_id= channel_id;
  data.value= * (ecdh->value());
  //  cerr << "Send data <<" << currtime() << endl;
  msource.start_message();
  default_ec_dii_factory->marshal( msource,
				   data);
  msource.end_message();
  //  cerr << "Send data >>" << currtime() << endl;
}


Forwarder_i::
~Forwarder_i() {
  omni_mutex_lock oml(guard);
  trace.start(100) << "~Forwarder_i" << '\n';
  delete &msource;
  delete []unique_description;
}

//------------------------------------------------------------McastMaster_i--
McastMaster_i::
McastMaster_i( McastSourceFactory &_msource_factory,
	       CosNaming::NamingContext_ptr root_ns,
	       const char *context_name,
	       unsigned long _slave_keep_alive) :
  msource_factory( _msource_factory),
  worker_condition( &worker_guard),
  slave_keep_alive( _slave_keep_alive)
{

  CONNECT_IMPL(Bootstrap::boa_ptr, this);

  CosNaming::Name ncontext= Bootstrap::create_name(context_name);
  CosNaming::Name nm( ncontext);
  nm.length( nm.length()+1);
  nm[nm.length()-1].id= 
    nm[nm.length()-1].kind= (const char *) "Master";
  try {
    root_ns->unbind( nm);
  }
  catch (...) {}
  try {
    Bootstrap::incremental_bind( root_ns, nm, this);
  }
  catch (CORBA::Exception &ex) {
    cerr << "Could not register ECMcast::Master reference in NS" << endl;
    cerr << ex << endl;
    exit(1);
  }

  (new omni_thread(static_worker,this))->start();
}

McastMaster_i::
~McastMaster_i() {
  //  cerr << "~McastMaster_i" << endl;
}

void McastMaster_i::
static_worker( void *args) {
  ((McastMaster_i*)args)->worker();
}

void McastMaster_i::
worker() {
  if (!slave_keep_alive) return;

  for (;;) { // for...ever
    ECMcast::Slaves slaves_copy;
    
    {
      GeneralLock::auto_read gard(guard);
      slaves_copy.length( slaves.size());
      CORBA::ULong i=0;
      for (MapSlaves::iterator it( slaves.begin());
	   it != slaves.end();
	   it++)
	slaves_copy[i++]= ECMcast::Slave::_duplicate( (*it).first);    
    }

    for (CORBA::ULong i=0; i<slaves_copy.length(); i++) {
      int dead= 1;
      try {
	dead= slaves_copy[i]->_non_existent();
      }
      catch (...) {
      }
      if (dead) {
	omni_mutex_lock oml( worker_guard);
	dead_slaves.push( ECMcast::Slave::_duplicate( slaves_copy[i]));
      }			
    }
    
    unsigned long sec,nsec;
    omni_thread::get_time( &sec,&nsec, 
			   slave_keep_alive, 0); // parameterize!, done
    int tw= 0;
    do {
      for (;;) {
	ECMcast::Slave_var a_slave;
	{
	  omni_mutex_lock oml(worker_guard);
	  if (dead_slaves.empty()) break;
	  a_slave= dead_slaves.front();
	  dead_slaves.pop();
	}
	trace.start(100) << "slave failed" << '\n';
	slave_failed( a_slave);
      }
      {
	omni_mutex_lock oml(worker_guard);
	tw= worker_condition.timedwait( sec, nsec);
      }
    } while (tw); // wait until timeout expires
  }
}

void McastMaster_i::
signal_failed_slave( ECMcast::Slave_ptr slave) {

  omni_mutex_lock oml(worker_guard);
  dead_slaves.push( ECMcast::Slave::_duplicate( slave));
  worker_condition.signal();
}

void McastMaster_i::
    slave_failed( ECMcast::Slave_ptr a_slave ) {
  
  MapForwarders copy_forwarders;
  {
    // first; remove slave from list of valid slaves, so it won't
    // be used by get_slave method
    //
    GeneralLock::auto_write gawl( guard);
    slaves.erase( a_slave);
    copy_forwarders= forwarders;
  }
  
  // inform all forwarders that a slave has gone down
  //
  for (MapForwarders::iterator it( copy_forwarders.begin());
       it != copy_forwarders.end();
       it ++) {

    ((Forwarder_i*)(*((*it).second)))->slave_failed( a_slave);
  }
}

AbstractForwarderInstance McastMaster_i::
create_forwarder() {
  AbstractForwarderInstance ret=
    AbstractForwarderInstance( 
		  new Forwarder_i( *this, 
				   *(msource_factory.create(NULL,0))),
		  0);
  return ret;
}

ECMcast::Slave_ptr McastMaster_i::
get_slave( CosEventComm::PushConsumer_ptr push_cons) {
  GeneralLock::auto_read gard(guard);

  ECMcast::Slave_var slave= NULL;
  CORBA::Short priority= 0;

  CORBA::String_var str= Bootstrap::orb_ptr->object_to_string( push_cons);

  IORProfiles iorp;
  if (IOR2Profiles( str, iorp)) {
    trace.start(400) << "IOR decoded" << '\n';
    for (CORBA::ULong i=0; i<iorp.profiles_length; i++) {
      if ((const char *) iorp.profiles[i].host) {
	trace.start(400) << iorp.profiles[i].host << '\n';
	
	hostent *h= gethostbyname( (const char *) iorp.profiles[i].host);
	if (h== NULL) return NULL;
	CORBA::Octet *addr= (CORBA::Octet*)h->h_addr;
	for (MapSlaves::iterator it( slaves.begin());
	     it != slaves.end();
	     it++) {

	  ECMcast::SlaveAddresses &sa= (*it).second;
	  for (CORBA::ULong x=0; x<sa.length(); x++) {
	    if (sa[x].priority < priority) continue;
	    CORBA::ULong v;
	    for (v=0; v<4; v++) {
	      trace.start(400) << "Compare "
			       << (unsigned long) sa[x].ipv4_address[v] << ','
			       << (unsigned long) addr[v] << '\n';
	      if (sa[x].ipv4_address[v] != addr[v]) break;
	    }
	    if (v==4) {
	      slave= ECMcast::Slave::_duplicate( (*it).first);
	      priority= sa[x].priority;
	    }		  
	  }
	}	
	break;
      }
    }
  }
  return ECMcast::Slave::_duplicate( slave);
}

void McastMaster_i::
register_slave( ECMcast::Slave_ptr new_slave,
		const ECMcast::SlaveAddresses &addresses) {

  trace.start(100) << "register_slave" << '\n';
  GeneralLock::auto_write gawl(guard);
  MapSlaves::iterator it=
    slaves.find(new_slave);
  slaves[it==slaves.end() ? ECMcast::Slave::_duplicate( new_slave) :
	new_slave]= addresses;
}

ECMcast::Time McastMaster_i::
get_current_time() {
  ECMcast::Time ret;

  TimeStamp ts= default_tsf->instantiate();
  ret.seconds= ts.sec;
  ret.microseconds= ts.nsec/1000;
  return ret;
}

void McastMaster_i::
consumer_failed( ECMcast::ChannelID channel_id,
		 ECMcast::ConsumerID consumer_id) {
  
  trace.start(500) << "consumer_failed" << '\n';
  ECRef holder;
  {
    GeneralLock::auto_write gawl(guard);
    MapForwarders::iterator it=
      forwarders.find(channel_id);
    if (it == forwarders.end()) return;
    holder= ECRef( *(((Forwarder_i*)(* (*it).second))->get_ec()));
  }
  CosEventChannelAdmin::ProxyPushSupplier_var proxy=
    holder.ec_i->find_proxy_push_supplier( consumer_id);
  if (!CORBA::is_nil(proxy)) {
    try {
      proxy->disconnect_push_supplier();
    }
    catch (...) {
    }
  }
}

#endif
