// $Id: master_forwarder.hh,v 1.2 1999/05/04 19:29:55 mencnar Exp $
//
// $Source: /home/iris/mencnar/ec/CVS/src/cc/master_forwarder.hh,v $
//
// $Log: master_forwarder.hh,v $
// Revision 1.2  1999/05/04 19:29:55  mencnar
// Added a method for synchronizing clocks.
//
// Revision 1.1  1999/02/24 13:01:58  mencnar
// *** empty log message ***
//
//


#include <ecmcast.hh>
#include <ecadmin.hh>
#include <iostream.h>
#include <stdlib.h>
#include "compatibility.hh"
#include "bootstrap.hh"
#include "locks.hh"
#include "ec.hh"
#include "mcast.hh"

#include <map>
#include <queue>
#include <deque>

class McastMaster_i;

#ifndef NO_MCAST

struct ForwardedProxy {
  ECMcast::Slave_var slave;
  CosEventChannelAdmin::ProxyPushSupplier_var proxy;
  int valid;

  ForwardedProxy() : valid(0) {}

  ForwardedProxy( ECMcast::Slave_ptr _slave,
		  CosEventChannelAdmin::ProxyPushSupplier_ptr _proxy) :
    slave( ECMcast::Slave::_duplicate( _slave)),
    proxy( CosEventChannelAdmin::ProxyPushSupplier::_duplicate( _proxy)),
    valid( 1)
  {
  }
};

class Forwarder_i : public AbstractForwarder {
  omni_mutex guard;
  typedef XSTD::map<CORBA::ULong,ForwardedProxy,
                    XSTD::less<CORBA::ULong> > Forwards;
  Forwards forwards;
  EventChannel_i *ec_i;
  CORBA::ULong channel_id;
  McastMaster_i &master;
  McastSource &msource;
  CORBA::ULong max_queue_size;
  ECAdmin::OverflowAction overflow_action;
  CORBA::ULong keep_alive_time;
  char *unique_description;
public:
  Forwarder_i( McastMaster_i &_master,
	       McastSource &_msource) :
    master( _master),
    msource( _msource),
    unique_description( _msource.get_unique_description()) {}

  virtual void
  open_event_channel( EventChannel_i &channel,
		      AbstractForwarderInstance inst,
		      CORBA::ULong _channel_id,
		      CORBA::ULong _max_queue_size,
		      ECAdmin::OverflowAction _oa,
		      CORBA::ULong _keep_alive_time);

  virtual int
  try_forward_consumer( CosEventChannelAdmin::ProxyPushSupplier_ptr push_supp,
		        CORBA::ULong consumer_id,
			CosEventComm::PushConsumer_ptr push_cons);

  virtual int
  try_forward_typed_consumer( CosEventChannelAdmin::ProxyPushSupplier_ptr push_supp,
			      CORBA::ULong consumer_id,
			      IRMOD InterfaceDef_ptr idef,
			      CosEventComm::PushConsumer_ptr cons_cons);
			   
  virtual void 
  remove_forwarded_consumer( CORBA::ULong consumer_id);
  
  virtual void 
  close_event_channel();
  // called by event channel from operation destroy,
  // AbstractForwarder should destroy all references of this event channel
  //

  virtual void
  push( ECDHandler &ecdh);

  EventChannel_i * get_ec() const { return ec_i; }
  void slave_failed( ECMcast::Slave_ptr a_slave);

  virtual ~Forwarder_i();
};
#endif // NO_MCAST

class McastMaster_i
  : public BOA_IMPL_BASE(ECMcast,Master) 
{
#ifndef NO_MCAST
  typedef XSTD::map<ECMcast::Slave_ptr,ECMcast::SlaveAddresses,
                    XSTD::less<ECMcast::Slave_ptr> > MapSlaves;
  MapSlaves slaves;

  typedef XSTD::queue<ECMcast::Slave_ptr,
                      XSTD::deque<ECMcast::Slave_ptr> > DeadSlaves;

  DeadSlaves dead_slaves;
  omni_mutex worker_guard;
  omni_condition worker_condition;
  unsigned long slave_keep_alive;

  typedef XSTD::map<ECMcast::ChannelID,AbstractForwarderInstance,
                    XSTD::less<ECMcast::ChannelID> > MapForwarders;

  MapForwarders forwarders;
  McastSourceFactory &msource_factory;
  GeneralLock guard;

  static void static_worker( void *args);
  void worker();
  
 public:
  McastMaster_i( McastSourceFactory &_msource_factory,
		 CosNaming::NamingContext_ptr root_ns,
		 const char *context_name,
		 unsigned long _slave_keep_alive);
  ~McastMaster_i();

  AbstractForwarderInstance create_forwarder();

  void insert_forwarder( CORBA::ULong channel_id,
			 AbstractForwarderInstance forw) {
    GeneralLock::auto_write gawl(guard);
    forwarders[channel_id]= forw;
  }
  
  void remove_forwarder( CORBA::ULong channel_id) {
    GeneralLock::auto_write gawl(guard);
    forwarders.erase( channel_id);
  }

  ECMcast::Slave_ptr get_slave( CosEventComm::PushConsumer_ptr push_cons);

  void register_slave( ECMcast::Slave_ptr new_slave,
		       const ECMcast::SlaveAddresses &addresses
		       ENV_DECL(env));
  void slave_failed( ECMcast::Slave_ptr a_slave
		     ENV_DECL(env));

  void signal_failed_slave( ECMcast::Slave_ptr slave);
  // called by one of the Forwarder_i objects when an exception
  // was caught while contacting a slave
  //

  ECMcast::Time get_current_time();

  void consumer_failed( ECMcast::ChannelID channel_id,
			ECMcast::ConsumerID consumer_id
			ENV_DECL(env));
#else
 public:
  virtual AbstractForwarderInstance create_forwarder()= 0;

#endif  // NO_MCAST
};


