// $Id: slave_daemon.cc,v 1.3 1999/05/21 16:35:47 mencnar Exp $
//
// $Source: /home/iris/mencnar/ec/CVS/src/cc/slave_daemon.cc,v $
//
// $Log: slave_daemon.cc,v $
// Revision 1.3  1999/05/21 16:35:47  mencnar
// Fixed displayed help information
//
// Revision 1.2  1999/05/04 19:37:34  mencnar
// Added QoS parameters, keep alive calls now performed by a slave process
//
// Revision 1.1  1999/02/24 13:02:00  mencnar
// *** empty log message ***
//
//

#ifndef NO_MCAST

#include <iostream.h>
#include "slave_daemon.hh"

#include "IRread.hh"
#include "bootstrap.hh"
#include "compatibility.hh"
#include "trace.hh"
#include "stream_trace.hh"
#include "mtime.hh"


#ifdef __WIN32__
#include <windows.h>
typedef int socklen_t;
#else
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#endif

#ifdef __sunos__
extern "C" int gethostname(char *name, int namelen);
#endif

SlaveForwarder::
SlaveForwarder( McastSlave_i *_mslave,
		CosEventComm::PushConsumer_ptr push_cons,
		ECMcast::ChannelID _channel_id,
		ECMcast::ConsumerID _consumer_id,		  
		CORBA::ULong _max_queue,
		ECAdmin::OverflowAction _overflow_action,
		CORBA::ULong _keep_alive_time) :
  failed_flag( 0),
  mslave( _mslave),
  channel_id( _channel_id),
  consumer_id( _consumer_id),
  queue_not_empty( &guard),
  max_queue( _max_queue),
  overflow_action( _overflow_action),
  keep_alive_time( _keep_alive_time),
  thr( new omni_thread( static_worker, this)),
  push_consumer( CosEventComm::PushConsumer::_duplicate( push_cons)),
  active( 1),
  is_typed( 0),
  ec_dii( NULL)
{
  thr->start();
}

SlaveForwarder::
SlaveForwarder( McastSlave_i *_mslave,
		CosEventComm::PushConsumer_ptr push_cons,
		IRMOD InterfaceDef_ptr _idef,
		ECMcast::ChannelID _channel_id,
		ECMcast::ConsumerID _consumer_id,		  
		CORBA::ULong _max_queue,
		ECAdmin::OverflowAction _overflow_action,
		CORBA::ULong _keep_alive_time) :
  failed_flag( 0),
  mslave( _mslave),
  channel_id( _channel_id),
  consumer_id( _consumer_id),
  queue_not_empty( &guard),
  max_queue( _max_queue),
  overflow_action( _overflow_action),
  keep_alive_time( _keep_alive_time),
  thr( new omni_thread( static_worker, this)),
  push_consumer( CosEventComm::PushConsumer::_duplicate( push_cons)),
  active( 1),
  is_typed( 1),
  ec_dii( NULL)
{ 
  try {
    if (!check_interface( _idef, tpintf)) {
      cerr << "Interface "  
	   << " does not fulfil contraints for typed event channel" << endl;
      throw 1;
    }
  }
  catch (CORBA::Exception &ex) {
    cerr << ex << endl;
    throw 1;
  }
  thr->start();
}

SlaveForwarder::
~SlaveForwarder() {
  if (ec_dii) delete ec_dii;
  trace.start(100) << "~SlaveForwarder\n";
}

void SlaveForwarder::
stop() {
  omni_mutex_lock oml(guard);
  active= 0;
  queue_not_empty.signal();  
}

void SlaveForwarder::
static_worker( void *args) {
  SlaveForwarder &sl= (*(SlaveForwarder*)args);
  sl.worker();
  sl.mslave->slave_finished( ComplexID( sl.channel_id, sl.consumer_id));
  delete &sl;
}

void SlaveForwarder::
worker() {
  if (is_typed) {
    try {
      CosTypedEventComm::TypedPushConsumer_var tpc=
	CosTypedEventComm::TypedPushConsumer::_narrow( push_consumer);
      
      if (!CORBA::is_nil(tpc)) 
	typed_consumer= tpc->get_typed_consumer();
      if (!CORBA::is_nil(typed_consumer)) {
	ec_dii= default_ec_dii_factory->create_ec_dii( typed_consumer,
						       tpintf);
      }
    }
    catch (...) {
      typed_consumer= NULL;
    }
    if (CORBA::is_nil(typed_consumer)) {
      trace.start(100) << "Unable to obtain typed push consumer" << '\n';
    }
  }

  ECDHandler ecdh;
  for (;;) {
    {
      {
	omni_mutex_lock oml(guard);
	if (!active) goto out;
	while (equeue.empty()) {
	  unsigned long sleep_s, sleep_ns;
	  omni_thread::get_time( &sleep_s, &sleep_ns, 
				 keep_alive_time, 0);
	  if (keep_alive_time) queue_not_empty.timedwait(sleep_s,sleep_ns);
	  else queue_not_empty.wait();

	  if (!active) goto out;
	  if (equeue.empty()) {
	    if (!keep_alive_time) continue;
	    guard.unlock();
	    int is_dead= 1;
	    try {
	      is_dead= push_consumer->_non_existent();
	    }
	    catch (CORBA::BAD_OPERATION &) {
	      is_dead= 0;
	    }
	    catch (...) {
	    }
	    guard.lock();
	    if (is_dead) goto fail;
	    continue;
	  }
	}
	ecdh= equeue.back();
	equeue.pop_back();
      }
      if (is_typed) {
	//	cerr << "typed" << endl;
	if (ecdh->value()->_d() == ECHelper::DT_ANY) {
	  try {
	    push_consumer->push( ecdh->value()->any_value());
	  }
	  catch (...) {
	    // consumer has failed
	    goto fail;
	  }
	}
	else {
	  //	  cerr << "typed data" << endl;
	  // forward typed data
	  if (!CORBA::is_nil(typed_consumer)) {
	    //	    cerr << "typed data & consumer" << endl;
	    try {
	      ECHelper::TypedOperation &to= ecdh->value()->oper();
	      ec_dii->send( to.name, to.params);
	    }
	    catch (...) {
	      // consumer has failed
	      goto fail;
	    }
	  }
	}
      }
      else {
	//	cerr << "untyped" << endl;
	//	cerr << ecdh->value() << endl;
	if (ecdh->value()->_d() == ECHelper::DT_ANY) {
	  try {
	    push_consumer->push( ecdh->value()->any_value());
	  }
	  catch (...) {
	    // consumer has failed
	    goto fail;
	  }
	}
	else {
	  assert(0);
	  // 
	  // it is not possible, a serious program failure
	  //
	}
      }
    }    
  }

 fail:
  failed_flag= 1;

 out:
  stop();

  // we disconnected a proxy, didn't we ?
  // it would be nice to call (or at least try to) a consumer 
  // about this situation
  //
  try {
    push_consumer->disconnect_push_consumer();
  }
  catch (...) {}
}

void SlaveForwarder::
push( ECDHandler &ecdh) {
  //  cerr << "SlaveForwarder::push" << endl;
  omni_mutex_lock oml(guard);
  if (!active) return;
  if (equeue.size() >= max_queue) {
    //
    // event queue overflow
    //
    switch (overflow_action) {
    case ECAdmin::OFA_DisconnectConsumer:
      try {
	active= 0;
	queue_not_empty.signal();  
	return;
      }
      catch (...) {}
      break;
    case ECAdmin::OFA_DropOldestMessage:
      if (!equeue.empty())
	equeue.pop_back();
      break;
    case ECAdmin::OFA_DropNewMessage:
      return;
    }
  }
  equeue.push_front(ecdh);
  queue_not_empty.signal();
}

//-------------------------------------------------------------McastSlave_i--
//

McastSlave_i::
McastSlave_i( McastSinkFactory &_sinkfact,
	      ECMcast::Master_ptr _master) :
  sinkfact( _sinkfact),
  master( ECMcast::Master::_duplicate( _master))
{
}

McastSlave_i::
~McastSlave_i() {
  trace.start(100) << "~McastSlave_i" << '\n';
}

void McastSlave_i::
static_worker( void *args) {
  WorkerArgs *wa= (WorkerArgs*) args;
  wa->mcslave->worker( *(wa->sink));
  delete wa;
}

// this worker keeps reading messages from a multicast sink and pushes
// them to an appropriate push consumer forwarder
//
void McastSlave_i::
worker( McastSink &sink) {
  while (1) {
    ECMcast::Data data;
    //    cerr << "call unmarshal" << endl;
    default_ec_dii_factory->unmarshal( sink, data);
    //    cerr << "Data received " << currtime() << endl;
    //    cerr << "unmarshal succeeded" << endl;
    ECHelper::ECValue *v= new ECHelper::ECValue( data.value);
    ECDHandler ecdh( new ECData(v), 0);

    omni_mutex_lock oml(guard);
    for (MapChannels::iterator it( channels.begin());
	 it != channels.end();
	 it++)
      if ((*it).first.channel_id == data.channel_id) {
	//	cerr << (void *) ((*it).second) << endl;
	(*it).second->push( ecdh);
      }
    //    cerr << ">>push" << endl;
  } 
}

CORBA::Boolean McastSlave_i::
forward_push_consumer( const char * mcast_channel_description,
		       ECMcast::ChannelID channel_id,
		       ECMcast::ConsumerID consumer_id,
		       CosEventChannelAdmin::ProxyPushSupplier_ptr supp,
		       CosEventComm::PushConsumer_ptr push_cons,
		       CORBA::ULong max_queue_size,
		       ECAdmin::OverflowAction oa,
		       CORBA::ULong keep_alive_time
		       ENV_DEF(env) ) {

  trace.start(100) << "forward_push_consumer" << '\n';
  omni_mutex_lock oml(guard);
  MapChannels::iterator it=
    channels.find( ComplexID( channel_id, consumer_id));
  if (it!=channels.end()) return (CORBA::Boolean) 1; 
  // cannot create forwarder once more

  McastSink *sink= sinkfact.create( mcast_channel_description);
  if (sinkset.find(sink) == sinkset.end()) {
    sinkset.insert(sink);
    (new omni_thread( static_worker, new WorkerArgs( this, sink)))->start();
  }

  channels[ComplexID(channel_id, consumer_id)]=
    new SlaveForwarder( this,
			push_cons,
			channel_id,
			consumer_id,		  
			max_queue_size,
			oa,
			keep_alive_time);
  return (CORBA::Boolean) 1;
}

CORBA::Boolean McastSlave_i::
forward_typed_push_consumer( const char * mcast_channel_description,
			     ECMcast::ChannelID channel_id,
			     ECMcast::ConsumerID consumer_id,
			     IRMOD InterfaceDef_ptr idef,
			     CosEventChannelAdmin::ProxyPushSupplier_ptr supp,
			     CosEventComm::PushConsumer_ptr push_cons,
			     CORBA::ULong max_queue_size,
			     ECAdmin::OverflowAction oa,
			     CORBA::ULong keep_alive_time
			     ENV_DEF(env) ) {

  trace.start(100) << "forward_typed_push_consumer" << '\n';
  omni_mutex_lock oml(guard);
  MapChannels::iterator it=
    channels.find( ComplexID( channel_id, consumer_id));
  if (it!=channels.end()) return (CORBA::Boolean) 1; 
  // cannot create forwarder once more

  McastSink *sink= sinkfact.create( mcast_channel_description);
  if (sinkset.find(sink) == sinkset.end()) {
    sinkset.insert(sink);
    (new omni_thread( static_worker, new WorkerArgs( this, sink)))->start();
  }

  channels[ComplexID(channel_id, consumer_id)]=
    new SlaveForwarder( this,
			push_cons,
			idef,
			channel_id,
			consumer_id,		  
			max_queue_size,
			oa,
			keep_alive_time);
  return (CORBA::Boolean) 1;
}

void McastSlave_i::
slave_finished( ComplexID cid) {
  // this call is always made from SlaveForwarded thread that it is just
  // finishing
  //
  trace.start(100) << "slave_finished" << '\n';
  SlaveForwarder *sf= NULL;
  {
    omni_mutex_lock oml(guard);
    MapChannels::iterator it= channels.find( cid);
    if (it!=channels.end()) {
      sf= (*it).second;
      channels.erase(it);
    }
  }
  if (sf) if (sf->failed()) {
    try {
      master->consumer_failed( cid.channel_id, cid.consumer_id);
    }
    catch (...) {
    }
  }
  trace.start(100) << "slave_finished: accepted" << '\n';
}

void McastSlave_i::
close_proxy( ECMcast::ChannelID channel_id,
	     ECMcast::ConsumerID consumer_id,
	     CORBA::Boolean call_consumer
	     ENV_DEF(env) ) {

  trace.start(100) << "close_proxy" << '\n';
  omni_mutex_lock oml(guard);
  MapChannels::iterator it=
    channels.find( ComplexID( channel_id, consumer_id));
  if (it==channels.end()) return; // a misinformed master ?

  (*it).second->stop();
  // initialize SlaveForwarder shutdown
}

void McastSlave_i::
master_restarted( ECMcast::Master_ptr new_master
		  ENV_DEF(env) ) {
  // not implemented, probably will never be used
  //
}

ECMcast::SlaveAddresses * McastSlave_i::
get_addresses() {
  ECMcast::SlaveAddresses *ret=
    new ECMcast::SlaveAddresses(1);
  ret->length(1);
  
  hostent *hp= NULL;
  char my_host_name[128];
  if (gethostname(my_host_name,sizeof(my_host_name))) {
    perror("Cannot get host name");
    exit(-1);
  }
  trace.start(10) << my_host_name << '\n';
  hp= gethostbyname(my_host_name);
  if (!hp) {
    perror("Cannot get local host address");
    exit(-1);
  }

  CORBA::Octet *addr= (CORBA::Octet*)hp->h_addr;
  (*ret)[0].priority= 100; 
  // this is my own address, so it would be best to forward all channels
  // through it
  (*ret)[0].ipv4_address.length(4);
  for (CORBA::ULong i=0; i<4; i++)
    (*ret)[0].ipv4_address[i]= addr[i];
  return ret;
}

void usage( const char *progname) {
  cerr << "Usage: " << progname << endl
       << "\t-context name" << endl
       << "\t-trace_level level" << endl;
  exit(1);
}

int main( int argc, char **argv) {

  StreamTrace strace( cerr);
  trace.set_abstract_trace( &strace);
  trace.set_trace_level( 0);

  // Initialize an ORB and Naming Service references
  //
  Bootstrap::init(argc,argv);
  CosNaming::NamingContext_var nsroot= Bootstrap::getNS();

  try {
    default_mcast_factory->init( argc, argv);
  }
  catch (McastFactory::InitializationError &ie) {
    cerr << "Error initializing multicast communcation layer" << endl;
    exit(1);
  }

  const char *context_name= NULL;
  for (char **pargv= argv+1; *pargv; pargv++) {
    if (!strcmp(*pargv, "-context") && *(pargv+1)) {
      context_name= *(++pargv);
    }
    else if (!strcmp(*pargv,"-h") || !strcmp(*pargv,"-?")) {
      usage(*argv);
    }
    else if (!strcmp(*pargv,"-trace_level") && *(pargv+1)) {
      trace.set_trace_level( atoi(*(++pargv)));
    }
    else {
      cerr << "Unknown option " << *pargv << endl;
      usage(*argv);
    }
  }
  if (!context_name) {
    cerr << "Undefined context name, please use \"-context\" option" << endl;
    usage(*argv);
  }

  // let's find the master object of the event channel factory,
  // we will need it, to register our services there
  //
  ECMcast::Master_var master;
  try {
    CosNaming::Name ctxnm( Bootstrap::create_name( context_name));
    CosNaming::Name master_name( ctxnm);
    master_name.length( master_name.length()+1);
    master_name[ master_name.length()-1].id=
      master_name[ master_name.length()-1].kind=
      (const char *) "Master";
    CORBA::Object_var obj= nsroot->resolve( master_name);
    master= ECMcast::Master::_narrow( obj);
    if (CORBA::is_nil(master)) throw 1;
  }
  catch (...) {
    cerr << "Could not find \"Master\" object of the event channel factory" << endl;
    return 1;
  }


  
  synch_time( 300, master);
  sleep(90);
  synch_time( 300, master);

  //  cerr << currtime() << endl;
  // well then,
  // so master is alive, let's inititialize multicast layer, and then
  // register using master interface
  //

  McastSinkFactory *mcast_sinkf;
  try {
    mcast_sinkf= default_mcast_factory->connect_as_slave();
  }
  catch (McastFactory::NoMasterActive &nma) {
    cerr << "The master of multicast communication group is not active" 
	 << endl;
    exit(1);
  }

  //
  //
  McastSlave_i *mcslave;
  try {
    mcslave= new McastSlave_i( *mcast_sinkf, master);
  }
  catch (McastFactory::NoMasterActive &nma) {
    cerr << "The master of multicast communication group is not active" 
	 << endl;
    exit(1);
  }

  CONNECT_IMPL(Bootstrap::boa_ptr,mcslave);
  try {
    ECMcast::SlaveAddresses_var sadds=
      mcslave->get_addresses();
    master->register_slave( mcslave, sadds);
  }
  catch (CORBA::Exception &ex) {
    Bootstrap::boa_ptr->dispose( mcslave);
    cerr << ex << endl;
    return 1;
  }

  trace.start(10) << "Implementation is ready for service" << '\n';
  IMPL_IS_READY(Bootstrap::boa_ptr);
  return 0;
}

#else 

#include <iostream.h>
int main( int argc, char **argv) {
  cerr << "Event channel daemon not supported for this configuration" << endl;
}

#endif // NO_MCAST

//




