// $Id: locks.hh,v 1.1.1.1 1998/12/13 23:27:16 mencda Exp $
//
// $Source: /home/mencda/CVSROOT/ec/src/hh/locks.hh,v $
//
// $Log: locks.hh,v $
// Revision 1.1.1.1  1998/12/13 23:27:16  mencda
// Imported sources
//
// Revision 1.1  1998/11/03 18:07:28  mencda
// Initial revision
//
//


#ifndef __locks_hh__
#define __locks_hh__

#include <omnithread.h>

class GeneralLock { // reader/writer reentrant lock
  omni_mutex guard;
  omni_condition waiting;
  int count;
  int writer_tid;
public:

  GeneralLock() :
    waiting( &guard),
    count( 0),
    writer_tid( 0) {}

  void LockRead() {
    omni_mutex_lock al(guard);
    if (count<0 && writer_tid == omni_thread::self()->id()) {
      // this thread holds writer's lock
      --count;
      return;
    }
    // another thread holds writer's lock
    while (count<0)
      waiting.wait();
    count++;
  }

  void UnlockRead() {
    omni_mutex_lock al(guard);
    if (count<0) {
      count++;
      return;
    }
    if (--count == 0) 
      waiting.signal(); // no threads waiting on this lock
  }

  class auto_read {
    GeneralLock &gl;
  public:
    auto_read( GeneralLock &_gl) : gl(_gl) {
      gl.LockRead();
    }
    ~auto_read() {
      gl.UnlockRead();
    }
  };

  void LockWrite() {
    omni_mutex_lock al(guard);
    int curr_tid= omni_thread::self()->id();
    if (count<0 && writer_tid == curr_tid) {
      --count;
      return;
    }
    while (count)
      waiting.wait();
    --count;
    writer_tid= curr_tid;
  }

  void UnlockWrite() {
    omni_mutex_lock al(guard);
    if (++count == 0)
      waiting.broadcast();
  }

  class auto_write {
    GeneralLock &gl;
  public:
    auto_write( GeneralLock &_gl) : gl(_gl) {
      gl.LockWrite();
    }
    ~auto_write() {
      gl.UnlockWrite();
    }
  };
};


#endif

// __locks_hh__


