// $Id: refcount.hh,v 1.2 1999/01/09 14:27:23 mencda Exp $
//
// $Source: /home/mencda/CVSROOT/ec/src/hh/refcount.hh,v $
//
// $Log: refcount.hh,v $
// Revision 1.2  1999/01/09 14:27:23  mencda
// *** empty log message ***
//
// Revision 1.1.1.1  1998/12/13 23:27:16  mencda
// Imported sources
//
//

#ifndef __REFCOUNT_HH__
#define __REFCOUNT_HH__

#include <assert.h>
#include <omnithread.h>
#include <iostream.h>
#include "utils.hh"

class DummyFriend {};

template <class T>
class RefCount {
  friend class DummyFriend; 
  // to keep GCC quiet "..only defines a private destructor and has no friends"

  omni_mutex guard;
  T * r;
  int ref_count;
public:

  RefCount(T *t) : 
    ref_count(1),
    r(t) 
  {
    //    cerr << "RefCount ctor " << (void*) this << " " << (void*) r << endl;
  }
  
  void duplicate() {
    omni_mutex_lock oml(guard);
    ++ref_count;
    //    cerr << "duplicate " << (void*) r << " " << ref_count << endl;
  }

  void release() {
    guard.lock();
    if (--ref_count == 0) {
      //      cerr << "release " << ref_count << endl;
      guard.unlock();
      delete this;
      return;
    }
    //    cerr << "release " << (void*) r << " " << ref_count << endl;
    guard.unlock();
  }

  T * val() {
    return r;
  }

private:
  ~RefCount() {
    //    cerr << "RefCount dtor " << (void*) this << endl;
    delete r;
  }

  DISABLE_COPY_AND_ASSIGN(RefCount)
};

template <class T>
class RefInstance {
  RefCount<T> *ref;
public:
  RefInstance() :
    ref( NULL) 
  {
    //    cerr << "default(0) RefInstance" << endl;
  }
  RefInstance( T *t, int ignored) :
    ref( new RefCount<T>(t)) 
  {
    //    cerr << "new RefInstance" << endl;
  }

  RefInstance( const RefInstance &n) :
    ref( n.ref) 
  {
    //    cerr << "copy ctor" << endl;
    if (ref) ref->duplicate();
    //    else cerr << "null object" << endl;
  }
  ~RefInstance() {
    if (ref) ref->release();
  }
  T * operator->() {
    return ref->val();
  }
  T * operator*() {
    return ref->val();
  }
  RefInstance & operator = ( const RefInstance &n) {
    //    cerr << "assign oper" << endl;
    if (this == &n || ref==n.ref) return *this;
    if (ref) ref->release();
    ref= n.ref;
    if (ref) ref->duplicate();
    //    else cerr << "null object" << endl;
    return *this;
  }

  // pointer comparison
  //
  int operator < (const RefInstance &src) const {
    return ref < src.ref;
  }

  int operator > (const RefInstance &src) const {
    return ref > src.ref;
  }

  int operator == (const RefInstance &src) const {
    return ref == src.ref;
  }

  int operator != (const RefInstance &src) const {
    return ref != src.ref;
  }
};


#endif // __REFCOUNT_HH__

//
