
//Title:        Event Channel Manager
//Version:      0.1
//Copyright:    Copyright (c) 1999
//Author:       Daniel Mencnarowski


package pl.edu.agh.ics.mev;

import org.omg.CORBA.*;
import org.omg.CosNaming.*;

import Repository;
import RepositoryHelper;

public class ConfigSet {

  // An initial reference resolved by:
  //
  public static final int InitRef_resolve= 0;  // calling resolve_initial_ref..
  public static final int InitRef_URL=     1;  // getting a reference from URL
  public static final int InitRef_IOR=     2;  // getting a reference from IOR
  public static final int InitRef_NS=      3;  // finding a reference in NS

  public NamingContext NS= null;
  public String NSResolve= null;
  public int NSResolveType= InitRef_resolve;

  public Repository IR= null;
  public String IRResolve= null;
  public int IRResolveType= InitRef_resolve;

  public org.omg.CORBA.ORB orb;

  private String readURL(String url) throws Exception {
    byte buf[]= new byte[4096];
    int len= 0;
    if (url.startsWith("file://")) {
      len= (new java.io.FileInputStream( url.substring(7, url.length()))).read(buf); 
    }
    else {
      java.net.URL neturl= new java.net.URL(url);
      len= neturl.openStream().read(buf);
    }
    while (len>0 && buf[len-1]<32)
      len--;
    return new String(buf, 0, len);
  }

  public static final org.omg.CosNaming.NameComponent[] getNSName( String s) {
    org.omg.CosNaming.NameComponent[] ret=
      new org.omg.CosNaming.NameComponent[1];
    ret[0]= new org.omg.CosNaming.NameComponent();
    ret[0].id= ret[0].kind= s;
    return ret;
  }

  public org.omg.CosNaming.NamingContext getNS() throws Exception {
    if (NS==null) resolveNS();
    return NS;
  }

  public Repository getIR() throws Exception {
    if (IR==null) resolveIR();
    return IR;
  }

  public void resolveNS() throws Exception {
    switch (NSResolveType) {
    case InitRef_resolve:
      NS= NamingContextHelper.narrow(orb.resolve_initial_references("NameService"));
      break;
    case InitRef_URL:
      NS= NamingContextHelper.narrow(orb.string_to_object(readURL(NSResolve)));
      break;
    case InitRef_IOR:
      NS= NamingContextHelper.narrow(orb.string_to_object(NSResolve));
      break;
    }
    if (NS==null) throw new Exception( "Operation failed");
    if (NS._non_existent()) throw new Exception( "Operation failed");
  }

  public void resolveIR() throws Exception {
    switch (IRResolveType) {
    case InitRef_resolve:
      IR= RepositoryHelper.narrow(orb.resolve_initial_references("InterfaceRepository"));
      break;
    case InitRef_URL:
      IR= RepositoryHelper.narrow(orb.string_to_object(readURL(IRResolve)));
      break;
    case InitRef_IOR:
      IR= RepositoryHelper.narrow(orb.string_to_object(IRResolve));
      break;
    case InitRef_NS:
      IR= RepositoryHelper.narrow(getNS().resolve(getNSName(IRResolve)));
      break;
    }
    if (IR==null) throw new Exception( "Operation failed");
    if (IR._non_existent()) throw new Exception( "Operation failed");
  }

  public ConfigSet( String []args, java.util.Properties[] pr) {
    orb= ORB.init( args, null);
    String prop= System.getProperty("NS_IOR_URL");
    if (prop != null) {
      NSResolve= prop;
      NSResolveType= InitRef_URL;
    }
    else if ((prop= System.getProperty("NS_IOR_FILE")) != null) {
      NSResolve= "file://"+prop;
      NSResolveType= InitRef_URL;
    }

    if ((prop= System.getProperty("IR_IOR_URL"))!= null) {
      IRResolve= prop;
      IRResolveType= InitRef_URL;
    }
    else if ((prop= System.getProperty("IR_IOR_FILE")) != null) {
      IRResolve= "file://"+prop;
      IRResolveType= InitRef_URL;
    }
    else if ((prop= System.getProperty("IR_NS")) != null) {
      IRResolve= prop;
      IRResolveType= InitRef_NS;
    }
  }
}
