// $Id$
//
// $Source$
//
// $Log$
//

#include "IORlib.hh"
#include <string.h>



char * Octets2IOR( CORBA::Octet *oct, CORBA::ULong dlen) {
  static const char hexstr[]= "0123456789ABCDEF";
  char *ret= CORBA::string_alloc( 1+ dlen*2+4);
  strcpy(ret,"IOR:");
  *(ret+dlen*2+4)= 0;
  char *ptr= ret+4;
  for (int i=0; i<dlen; i++) {
    *(ptr++)= hexstr[((oct[i] & 0xf0) >> 4)];
    *(ptr++)= hexstr[oct[i] & 0x0f];
  }
  return ret;
}

CORBA::Octet* IOR2Octets( const char *ior, CORBA::ULong &dlen) {
  int len= strlen( ior);
  if (len < 4 || len % 2) return NULL;
  if (strncmp( ior, "IOR:", 4)) return NULL;
  CORBA::Octet *ret= new CORBA::Octet[dlen= (len-4)/2];
  CORBA::Octet *retptr= ret;
  for ( const char *ptr= ior+4;
        *ptr;
        ptr+=2, retptr++) {
    unsigned v= 0;
    for (unsigned i=0; i<2; i++) {
      CORBA::Octet nv;
      if (*(ptr+i)>='0' && *(ptr+i)<='9') nv= (*(ptr+i) - '0');
      else if (*(ptr+i)>='a' && *(ptr+i)<='f') nv= (*(ptr+i) - 'a' +10);
      else if (*(ptr+i)>='A' && *(ptr+i)<='F') nv= (*(ptr+i) - 'A' +10);
      else {
        delete []ret;
        return NULL;
      }
      v+= (i ? 1 : 16)*nv;
    }
    (*retptr)= (CORBA::Octet) v;
  }
  return ret;
}



int IOR2Profiles( const char *str, IORProfiles &iorp) {
  CORBA::ULong alllen= 0;
  CORBA::ULong iptr= 0;
  CORBA::Octet* alldata= IOR2Octets( str, alllen);

  try {
    if (alldata==NULL) return 0;
    
    iorp.byte_order= iorp.get_char( alldata, iptr, alllen);
    //    cerr << "byte_order " << (unsigned) iorp.byte_order << endl;
    CORBA::ULong type_id_len= iorp.get_ulong( alldata, iptr, alllen);
    //    cerr << "type_id_len " << (unsigned) type_id_len << endl;
    if (type_id_len==0) 
      iorp.type_id= CORBA::string_dup("");
    else {
      //      cerr << iptr << endl;
      if (iptr+type_id_len>alllen) throw 1;
      if (alldata[iptr+type_id_len-1]) throw 1;
      char *s= CORBA::string_alloc( type_id_len);
      memcpy( s, &alldata[iptr], type_id_len);
      iorp.type_id= s;
      iptr+= type_id_len;
    }
    //    cerr << "type_id " << (const char *) iorp.type_id << endl;
    iorp.profiles_length= iorp.get_ulong( alldata, iptr, alllen);
    if (iorp.profiles) delete []iorp.profiles;
    iorp.profiles= new IORProfile[iorp.profiles_length];
    for (CORBA::ULong i=0; i<iorp.profiles_length; i++) {
      iorp.profiles[i].profileid= iorp.get_ulong( alldata, iptr, alllen);
      //      cerr << "profile_id " << (unsigned) iorp.profiles[i].profileid 
      //           << endl;
      CORBA::ULong profilelen=
	iorp.profiles[i].profile_length= 
	iorp.get_ulong( alldata, iptr, alllen);

      if (iptr+profilelen>alllen) throw 1;
      iorp.profiles[i].data= new CORBA::Octet[profilelen];
      memcpy(iorp.profiles[i].data, &alldata[iptr], profilelen);
      iptr+= profilelen;
      
      if (iorp.profiles[i].profileid == 0) { // IIOP profile
	CORBA::Octet *pdata= iorp.profiles[i].data;
	CORBA::ULong xptr= 0;
	CORBA::ULong xlen= profilelen;
	iorp.profiles[i].major_version= 
	  iorp.get_char(pdata,xptr,xlen);
	iorp.profiles[i].minor_version= 
	  iorp.get_char(pdata,xptr,xlen);
	//	cerr << "IIOP " << (unsigned) iorp.profiles[i].major_version
	//	     << "." << (unsigned) iorp.profiles[i].minor_version 
	//	     << endl;
	CORBA::ULong hostlen= iorp.get_ulong(pdata,xptr,xlen);
	if (xptr+hostlen>xlen) throw 1;
	if (pdata[xptr+hostlen-1]) throw 1;
	char *s= CORBA::string_alloc( hostlen);
	memcpy( s, &pdata[xptr], hostlen);
	iorp.profiles[i].host= s;
	xptr+= hostlen;
	//	cerr << "Host " << (const char *) iorp.profiles[i].host 
	//           << endl;
	iorp.profiles[i].port= iorp.get_ushort(pdata,xptr,xlen);
	//	cerr << "Port " << (unsigned) iorp.profiles[i].port << endl;
	CORBA::ULong key_length=
	  iorp.profiles[i].object_key_length=
	  iorp.get_ulong(pdata,xptr,xlen);
	if (xptr+key_length>xlen) throw 1;
	//	cerr << "Key length " 
	//           << (unsigned) iorp.profiles[i].object_key_length
	//	     << endl;
	iorp.profiles[i].object_key= new CORBA::Octet[key_length];
	memcpy( iorp.profiles[i].object_key,
		&pdata[xptr],
		key_length);
	xptr+= key_length;
      }
    } 
  }
  catch (...) {}
  
  delete []alldata;

  //  iorp.profiles= new IORProfile[1];
  //  iorp.profiles_length= 1;
  //  iorp.profiles[0].profileid= 0;
  //  iorp.profiles[0].host= CORBA::string_dup("peony");
  return 1;
}


