// $Id: bootstrap.cc,v 1.1.1.1 1998/12/13 23:27:16 mencda Exp $
//
// $Source: /home/mencda/CVSROOT/ec/src/lib/bootstrap.cc,v $
//
// $Log: bootstrap.cc,v $
// Revision 1.1.1.1  1998/12/13 23:27:16  mencda
// Imported sources
//
//

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>

#include "bootstrap.hh"
#include "compatibility.hh"

CosNaming::NamingContext_ptr Bootstrap::root_ns= NULL;
CORBA::ORB_ptr Bootstrap::orb_ptr= NULL;
CORBA::BOA_ptr Bootstrap::boa_ptr= NULL;

void Bootstrap::init( int &argc, char ** &argv) {
  orb_ptr= CORBA::ORB_init(argc,argv,ORB_IDENTIFIER);
  boa_ptr= orb_ptr->BOA_init(argc,argv,BOA_IDENTIFIER);

#ifdef OMNIORB
  // Turn off garbage collecting of unused connections
  //
  omniORB::idleConnectionScanPeriod( omniORB::idleIncoming, 0);
  omniORB::idleConnectionScanPeriod( omniORB::idleOutgoing, 0);  
#endif
}

void Bootstrap:: init( CORBA::ORB_ptr _orb_ptr,
		       CORBA::BOA_ptr _boa_ptr) {
  if (orb_ptr)
    cerr << "Bootstrap already initialized!" << endl;
  else {
    orb_ptr= _orb_ptr;
    boa_ptr= _boa_ptr;
  }
#ifdef OMNIORB
  // Turn off garbage collecting of unused connections
  //
  omniORB::idleConnectionScanPeriod( omniORB::idleIncoming, 0);
  omniORB::idleConnectionScanPeriod( omniORB::idleOutgoing, 0);  
#endif
}

static CosNaming::NamingContext_ptr myGetNS() {
  CORBA::Object_var nsobj;
  CosNaming::NamingContext_ptr ret= NULL;
  int nsmethod= 0;
  char *NSFILE= getenv("NS_IOR_FILE");
  if (NSFILE && strcmp(NSFILE,"-resolve_initial_references")) {
    ifstream ifs(NSFILE);
    if (!ifs) {
      cerr << "Could not open file " << NSFILE << " for reading" << endl;
      return ret;
    }
    char buf[2048];
    ifs >> buf;
    if (!ifs) {
      cerr << "Error reading file " << NSFILE << endl;
      return ret;
    }
    try {
      nsobj= Bootstrap::orb_ptr->string_to_object(buf);
      if (CORBA::is_nil(nsobj)) throw 1;
    }
    catch (...) {
      cerr << "Unable to convert " << NSFILE << " file" << endl
	   << " ($NS_IOR_FILE) into CORBA object" << endl;
      return ret;
    }
  }
  else {
    nsmethod= 1;
    try {
      nsobj= Bootstrap::orb_ptr->resolve_initial_references("NameService");
    }
    catch (...) {
      cerr << "Unable to get NameService reference through \"resolve_initial_references\"" << endl;
      return ret;
    }
  }
  try {
    ret= CosNaming::NamingContext::_narrow(nsobj);
  }
  catch (...) {
  }
  if (CORBA::is_nil(ret)) {
    cerr << "Failed to narrow NameService reference provided by "
	 << (nsmethod ? "\"resolve_initial_references\"" :
	     "($NS_IOR_FILE)") << endl;
    return ret;
  }  
  return ret;
}

CosNaming::NamingContext_ptr Bootstrap::getNS( int exit_on_failure) {
  if (!root_ns) {
    root_ns= myGetNS();
    if (CORBA::is_nil(root_ns) && exit_on_failure) exit(exit_on_failure);
  }
  return CosNaming::NamingContext::_duplicate(root_ns);
}

CosNaming::Name Bootstrap::SRV_IR() {
  CosNaming::Name ret(1);
  ret.length(1);
  ret[0].id= (const char *) "IR";
  ret[0].kind= (const char *) "IR";
  return ret;
}

CosNaming::Name Bootstrap::create_name( const char *_str) {
  char *str,*ostr;
  str= ostr= strdup(_str);
  int len=0;
  CosNaming::Name ret(len);

  char *beg= str;
  char c= '*';

  while (c) {
    while (*str && *str!='.' && *str!='-')
      str++;
    c= *str;
    *str= '\0';
    if (beg==str && !c) break;
    ret.length(++len);
    ret[len-1].id= 
      ret[len-1].kind= (const char *) beg;
    beg= ++str;

    if (c=='-') {
      while (*str && *str!='.' && *str!='-')
        str++;
     c= *str;
      *str= '\0';
      ret[len-1].kind= (const char *) beg;
      if (c=='\0') break;
      beg= ++str;
    }
  }
  free(ostr);
  return ret;
}

void Bootstrap::incremental_bind( CosNaming::NamingContext_ptr root_ns, 
				  const CosNaming::Name &name,
				  CORBA::Object_ptr object) {

  CosNaming::Name tmpName( name.length());

  for ( int i=1; i<name.length(); i++) {
    tmpName.length( i);

    for ( int j=0; j<i; j++)
      tmpName[ j]= name[ j];
    try {
      CosNaming::NamingContext_var nc( root_ns->bind_new_context( tmpName));
    }
    catch ( CosNaming::NamingContext::AlreadyBound &) {}
  }

  int len= name.length();
  tmpName.length( len);
  tmpName[ len-1]= name[ len-1];
  try {
    root_ns->unbind( tmpName);
  }
  catch (...) {
  }
  root_ns->bind( tmpName, object);
}

