// $Id: mtp.hh,v 1.1 1998/11/03 18:04:30 mencda Exp mencda $
//
// $Source: /home/mencda/thesis/mtp/mtp.hh,v $
//
// $Log: mtp.hh,v $
// Revision 1.1  1998/11/03 18:04:30  mencda
// Initial revision
//
//

#ifndef __MTP_HH__
#define __MTP_HH__

#include <iostream.h>
#include <trace.hh>

#include <map>
#include <set>
#include <vector>

#include <sys/types.h>

#ifdef __WIN32__
#include <windows.h>
#include <process.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#endif


#include "locks.hh"
#include "utils.hh"


#define MTP_PERMANENT_ADDR "224.0.1.9"

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

const u8 MTP_CURRENT_VERSION=        0x1;
const u32 MTP_UNKNOWN_ID=            0x0;

class MTPPacket;

struct Time {
  unsigned long sec;
  unsigned long nsec;

  Time() : sec(0L), nsec(0L) {}
  Time( unsigned long _sec,
	unsigned long _nsec) : sec(_sec), nsec(_nsec) {}

  Time& normalize() {
    while (nsec>=1000000000L) {
      nsec-= 1000000000L;
      ++sec;
    }
    return *this;
  }

  int operator < ( const Time &df) const {
    Time t1( *this);
    t1.normalize();
    Time t2( df);
    t2.normalize();
    return (t1.sec<t2.sec) ? 1 : 
      ((t1.sec>t2.sec) ? 0 :
       ((t1.nsec<t2.nsec) ? 1 : 0));
  }

  Time diff ( const Time &sb) const {
    Time t1(*this);
    Time t2(sb);
    if (*this < sb) {
      t1= sb;
      t2= *this;
    }
    t1.normalize();
    t2.normalize();
    Time ret;
    if (t1.nsec<t2.nsec) {
      ret.nsec= (t1.nsec+1000000000L)-t2.nsec;
      ++(t2.sec);
    }
    else ret.nsec= t1.nsec-t2.nsec;
    ret.sec= t1.sec-t2.sec;
    return ret;
  }
};


// a 48-bit wide unsigned number (used as message sequence number)
//
class u48 {
  u16 hi;
  u32 lo;
public:

  u48() : hi(0), lo(0) {}
  u48( u16 _hi, u32 _lo) : hi(_hi), lo(_lo) {}
  u48( const u48 &src) : hi(src.hi), lo(src.lo) {}
  u48 & operator =( const u48 &src) {
    hi= src.hi;
    lo= src.lo;
    return *this;
  }
  void reset() {
    hi= 0;
    lo= 0;
  }
  int operator == ( const u48 &src) const {
    return (hi == src.hi && lo == src.lo);
  }
  int operator < ( const u48 &src) const {
    if (hi<src.hi) return 1;
    if (hi>src.hi) return 0;
    return (lo<src.lo);
  }
  int operator <= ( const u48 &src) const {
    if (hi<src.hi) return 1;
    if (hi>src.hi) return 0;
    return (lo<=src.lo);
  }
  int operator > ( const u48 &src) const {
    return !(*this<=src);
  }
  int operator >= ( const u48 &src) const {
    return !(*this<src);
  }
  u48 operator + ( u32 x) {
    if (x==0) return *this;
    u48 ret(*this);
    if ((((u32) 4294967295UL) - x)+1 >= lo ) {
      int i=0;
      while ( ret.lo >= 1073741824UL && i<4) {
	ret.lo-= 1073741824UL;
	++i;
      }
      while ( x >= 1073741824UL && i<4) {
	x-= 1073741824UL;
	++i;
      }
      ret.lo+= x;
      ret.hi++;
    }
    else ret.lo+= x;
    return ret;
  }
  u16 get_hi() const { return hi; }
  u32 get_lo() const { return lo; }
  friend ostream & operator<<( ostream &, u48);
  friend TraceSink & operator<<( TraceSink &, u48);
};

inline ostream & operator<<( ostream &s, u48 v) {
  return s << v.lo;
}

inline TraceSink & operator<<( TraceSink &s, u48 v) {
  return s << v.lo;
}

// contains information about a single service access point and current
// parameters of the MTP web it is associated with
//
class MTPWeb {
  u32 r_heartbeat;
  u16 r_window;
  u16 r_retention;
  u32 r_mcast_connection_id;
  u32 r_master_connection_id;
  u32 r_ucast_connection_id;
  u16 r_packet_number;
  u48 r_sequence_number;
  u16 r_max_data_size;
  u16 r_reply_port;

public:
  MTPWeb() {}
  MTPWeb( u32 a_heartbeat,
	  u16 a_window,
	  u16 a_retention,
	  u16 a_max_data_size,
	  u32 a_ucast_connection_id,
	  u16 a_reply_port) : 
    // called by a client wishing to join a web
    r_heartbeat( a_heartbeat),
    r_window( a_window),
    r_retention( a_retention),
    r_mcast_connection_id( MTP_UNKNOWN_ID),
    r_master_connection_id( MTP_UNKNOWN_ID),
    r_ucast_connection_id( a_ucast_connection_id),
    r_packet_number( 0),
    r_max_data_size( a_max_data_size),
    r_reply_port( a_reply_port) {}
    
  MTPWeb( u32 a_heartbeat,
	  u16 a_window,
	  u16 a_retention,
	  u16 a_max_data_size,
	  u32 a_mcast_connection_id,
	  u32 a_master_connection_id,
	  u16 a_reply_port) : 
    // called by a master upon establishing a new web
    r_heartbeat( a_heartbeat),
    r_window( a_window),
    r_retention( a_retention),
    r_mcast_connection_id( a_mcast_connection_id),
    r_master_connection_id( a_master_connection_id),
    r_ucast_connection_id( a_master_connection_id),
    r_packet_number( 0),
    r_max_data_size( a_max_data_size),
    r_reply_port( a_reply_port) {}

  void update_after_join( const MTPPacket &);
    
  u32 heartbeat() const { return r_heartbeat; }
  // heartbeat (in milliseconds) 
  u16 window() const { return r_window; }
  // window size (a number of multicast packets transmitted by a single
  // member during a heartbeat)
  u16 retention() const { return r_retention; }
  // a number of heartbeats a sending client needs to retain data
  // to retransmit it if necessary
  u32 mcast_connection_id() const { return r_mcast_connection_id; }
  // a mutlicast connection id assigned by a master of a web
  // in join-confirm packet
  u32 master_connection_id() const { return r_master_connection_id; }
  // a unicast connection id of a master of the web
  u32 ucast_connection_id() const { return r_ucast_connection_id; }
  // a unicast connection id assigned locally and known to be unique
  // for a given machine

  void assign_packet_number() { ++r_packet_number; }
  void assign_sequence_number() { 
    r_sequence_number= r_sequence_number+1; 
    r_packet_number= 0;
  }
  void set_sequence_number( u48 seq_num) { 
    r_sequence_number= seq_num; 
    r_packet_number= 0;
  }
  u16 packet_number() const { return r_packet_number; }
  void set_packet_number( u16 packet_num) {
    r_packet_number= packet_num;
  }
  u48 sequence_number() const { return r_sequence_number; }
  u16 max_data_size() const { return r_max_data_size; }
  u16 reply_port() const { return r_reply_port; }
};

const unsigned MAX_MTP_PACKET_SIZE= 4096;

// Please note: sizeof(mtp_header) doesn't have to be equal to MTP_HEADER_SIZE
//


#
struct MTPHeader {
#if defined(__WIN32__) || defined(__sunos__)
  enum { 
    T_DATA=                0x00,
    T_DATA_DATA=           0x00,
    T_DATA_EOW=            0x01,
    T_DATA_EOM=            0x02,

    T_NAK=                 0x01,
    T_NAK_REQUEST=         0x00,
    T_NAK_DENY=            0x01,

    T_EMPTY=               0x02,
    T_EMPTY_DALLY=         0x00,
    T_EMPTY_CANCEL=        0x01,
    T_EMPTY_HIBERNATE=     0x02,
  
    T_JOIN=                0x03,
    T_JOIN_REQUEST=        0x00,
    T_JOIN_CONFIRM=        0x01,
    T_JOIN_DENY=           0x02,

    T_QUIT=                0x04,
    T_QUIT_REQUEST=        0x00,
    T_QUIT_CONFIRM=        0x01,

    T_TOKEN=               0x05,
    T_TOKEN_REQUEST=       0x00,
    T_TOKEN_CONFIRM=       0x01,

    T_IS_MEMBER=           0x06,
    T_IS_MEMBER_REQUEST=   0x00,
    T_IS_MEMBER_CONFIRM=   0x01,
    T_IS_MEMBER_DENY=      0x02,

    MEMBER_MASTER=         0x00,
    MEMBER_PRODUCER=       0x01,
    MEMBER_CONSUMER=       0x02
  };
#else
  static const u8 T_DATA=                0x00;
  static const u8 T_DATA_DATA=           0x00;
  static const u8 T_DATA_EOW=            0x01;
  static const u8 T_DATA_EOM=            0x02;

  static const u8 T_NAK=                 0x01;
  static const u8 T_NAK_REQUEST=         0x00;
  static const u8 T_NAK_DENY=            0x01;

  static const u8 T_EMPTY=               0x02;
  static const u8 T_EMPTY_DALLY=         0x00;
  static const u8 T_EMPTY_CANCEL=        0x01;
  static const u8 T_EMPTY_HIBERNATE=     0x02;
  
  static const u8 T_JOIN=                0x03;
  static const u8 T_JOIN_REQUEST=        0x00;
  static const u8 T_JOIN_CONFIRM=        0x01;
  static const u8 T_JOIN_DENY=           0x02;

  static const u8 T_QUIT=                0x04;
  static const u8 T_QUIT_REQUEST=        0x00;
  static const u8 T_QUIT_CONFIRM=        0x01;

  static const u8 T_TOKEN=               0x05;
  static const u8 T_TOKEN_REQUEST=       0x00;
  static const u8 T_TOKEN_CONFIRM=       0x01;

  static const u8 T_IS_MEMBER=           0x06;
  static const u8 T_IS_MEMBER_REQUEST=   0x00;
  static const u8 T_IS_MEMBER_CONFIRM=   0x01;
  static const u8 T_IS_MEMBER_DENY=      0x02;

  static const u8 MEMBER_MASTER=         0x00;
  static const u8 MEMBER_PRODUCER=       0x01;
  static const u8 MEMBER_CONSUMER=       0x02;
#endif

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_protocol_version =     0x00};
#else
  static const int POS_protocol_version =     0x00;
#endif
  u8 protocol_version;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_packet_type =          0x01};
#else
  static const int POS_packet_type =          0x01;
#endif
  u8 packet_type;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_packet_modifier =       0x02};
#else
  static const int POS_packet_modifier =       0x02;
#endif
  u8 packet_modifier;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_client_channel =        0x03};
#else
  static const int POS_client_channel =        0x03;
#endif
  u8 client_channel;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_src_id =                0x04};
#else
  static const int POS_src_id =                0x04;
#endif
  u32 src_id;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_dst_id =                0x08};
#else
  static const int POS_dst_id =                0x08;
#endif
  u32 dst_id;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_synchro =               0x0c};
#else
  static const int POS_synchro =               0x0c;
#endif
  u8 synchro;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_msg_state =             0x0d};
#else
  static const int POS_msg_state =             0x0d;
#endif
  u8 msg_state[3];

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_msg_seq_number =        0x12};
#else
  static const int POS_msg_seq_number =        0x12;
#endif
  u48 msg_seq_number;
  
#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_packet_seq_number =     0x10};
#else
  static const int POS_packet_seq_number =     0x10;
#endif
  u16 packet_seq_number;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_heartbeat =             0x18};
#else
  static const int POS_heartbeat =             0x18;
#endif
  u32 heartbeat;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_window =                0x2c};
#else
  static const int POS_window =                0x2c;
#endif
  u16 window;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_retention =             0x1e};
#else
  static const int POS_retention =             0x1e;
#endif
  u16 retention;

#if defined(__WIN32__) || defined(__sunos__)
  enum { POS_reply_port =             0x20};
#else
  static const int POS_reply_port =             0x20;
#endif
  u16 reply_port;

#if defined(__WIN32__) || defined(__sunos__)
  enum { MTP_HEADER_SIZE=       0x24};
#else
  static const unsigned MTP_HEADER_SIZE=       0x24;
#endif

  void swap_ids() {
    u32 tmp= src_id;
    src_id= dst_id;
    dst_id= tmp;
  }
};

struct PacketID {
  u48 message_no;
  u16 packet_no;
  PacketID( u48 m, u16 p) : message_no( m), packet_no( p) {}
  int operator < ( const PacketID &id) const {
    return (message_no < id.message_no ||
	    message_no == id.message_no &&
	    packet_no < id.packet_no);
  }
  int operator <= ( const PacketID &id) const {
    return (message_no < id.message_no ||
	    message_no == id.message_no &&
	    packet_no <= id.packet_no);
  }
  int operator > ( const PacketID &id) const {
    return (message_no > id.message_no ||
	    message_no == id.message_no &&
	    packet_no > id.packet_no);
  }
  int operator >= ( const PacketID &id) const {
    return (message_no > id.message_no ||
	    message_no == id.message_no &&
	    packet_no >= id.packet_no);
  }
  int operator == ( const PacketID &id) {
    return (message_no == id.message_no &&
	    packet_no == id.packet_no);
  }
};

struct TSAP { // transport service access point
  sockaddr addr;
  unsigned addr_len;
  u32 id;
  u16 reply_port;

  TSAP() {}
  TSAP( sockaddr an_addr, unsigned an_addr_len, u32 an_id) :
    addr( an_addr), addr_len( an_addr_len), id( an_id) {}

  int operator == ( const TSAP &tsap) const;
  int operator != ( const TSAP &tsap) const; // never defined
  int operator < ( const TSAP &) const; // never defined
  int operator > ( const TSAP &) const; // never defined
};

struct NAKRange {
public:
  NAKRange() {}
  
  u48 min_message;
  u16 min_packet;
  u48 max_message;
  u16 max_packet;
  
  int operator <( const NAKRange &) const { return 0; }
  int operator ==(const NAKRange &) const { return 0; }
  int operator >(const NAKRange &) const { return 0; }
};

class MTPPacket {
  unsigned char *all_data;
  void encode_header();

public:
  MTPPacket() :
    all_data( new unsigned char[MAX_MTP_PACKET_SIZE]),
    data( all_data+MTPHeader::MTP_HEADER_SIZE),
    data_length(0) {}
  
  MTPPacket( unsigned _data_length) :
    all_data( new unsigned char[MTPHeader::MTP_HEADER_SIZE+_data_length]),
    data( all_data+MTPHeader::MTP_HEADER_SIZE),
    data_length( _data_length) {}

  ~MTPPacket() {
    delete [] all_data;
  }

  MTPPacket( const MTPPacket &src);
  MTPPacket & operator=(const MTPPacket &);

  MTPHeader header;
  sockaddr addrfrom;
  unsigned addrfrom_len;
  unsigned char *data;
  unsigned data_length;

  int recv( int fd);
  int send( int fd, sockaddr &to, int tolen);
  int send_ucast_reply( int fd, unsigned port);

  void create_join_request( unsigned char member_class,
			    const MTPWeb &web);
  void create_join_confirm( const MTPWeb &web);
  void create_join_deny( const MTPWeb &web);

  void create_quit_request( const MTPWeb &web,
			    u32 dst_connection_id);
  void create_quit_confirm( const MTPWeb &web);
  
  void create_is_member_request( const MTPWeb &web,
				 u32 dst_connection_id);
  void create_is_member_confirm( const MTPWeb &web);
  void create_is_member_deny( const MTPWeb &web);
  
  void create_token_request( const MTPWeb &web);
  void create_token_confirm( const MTPWeb &web);
  
  void create_empty_packet( const MTPWeb &web,
			    u8 modifier 
			    // T_EMPTY_DALLY, T_EMPTY_CANCEL, T_EMPTY_HIBERNATE
			    );
  void create_data_packet( const MTPWeb &web,
			   u8 modifier,  
			   // T_DATA_DATA, T_DATA_EOM, T_DATA_EOW
			   unsigned a_data_length);
  
  unsigned create_nak_request( const MTPWeb &web,
			       u32 producer_connection_id,
			       XSTD::vector<NAKRange> &ranges);
  // returns number of nak request ranges consumed
  void create_nak_deny( const MTPWeb &web);

  int operator <( const MTPPacket &) const;
  int operator ==( const MTPPacket &) const;
};

int create_mcast_socket( const char *mcast_addr_name,
			 unsigned port,
			 int ttl);

int resolve_host( const char *hostname,
		  sockaddr &sa,
		  unsigned &sa_len,
		  unsigned port);



class IncomingPacketHandler {
public:
  virtual void handle( MTPPacket &packet)= 0;
};

class PacketEndpoint {
public:
  PacketEndpoint( unsigned _mtplistenport) :
    mtplistenport( _mtplistenport) {}

  u32 ipaddress() const { return ipaddr; }
  unsigned listen_port() const { return mtplistenport; }

  unsigned ipaddr;

protected:
  unsigned mtplistenport;

};

//-----------------------------------------------------------------PacketMgr--
//
class PacketMgr : public PacketEndpoint {
  int fd;
  int ucast_fd;
  unsigned reply_port;
  unsigned listen_port;
  char *address;
  int out_fd;
  sockaddr send_addr;
  unsigned send_addr_len;
  
  class INPHandler {
  public:
    INPHandler( PacketMgr &container, 
		INPHandler *next_free,
		INPHandler *next_busy,
		unsigned a_no) :
      mgr( container),
      next_free_handler( next_free),
      next_busy_handler( next_busy),
      no( a_no),
      thr( new omni_thread( handler, this)) {

      thr->start();
    }
    unsigned no;
    MTPPacket packet;
    PacketMgr &mgr;
    omni_thread *thr;
    static void handler( void *ptr);

    INPHandler *next_free_handler;
    INPHandler *next_busy_handler;

  } *free_handlers, *busy_handlers;

  friend class INPHandler;


public:
  struct MapItem {
    IncomingPacketHandler *iph;
    u32 my_connection_id;
    
    MapItem() : iph(0), my_connection_id(0) {}
    MapItem( IncomingPacketHandler *an_iph,
	     u32 a_connection_id) :
      iph( an_iph),
      my_connection_id( a_connection_id) {}
  };
  typedef XSTD::multimap<u32, MapItem, XSTD::less<u32> > map_type;

  PacketMgr( const char *mcast_address,
	     unsigned _mtplistenport,
	     int _ttl);
  ~PacketMgr();

  void register_handler( IncomingPacketHandler &hnd,
			 u32 ucast_connection_id,
			 u32 my_id);
  int unregister_handler( IncomingPacketHandler &hnd,
			  u32 ucast_connection_id);

  int get_fd() const { return fd; }
  int get_ucast_fd() const { return ucast_fd; }
  unsigned get_ucast_port() const { return reply_port; }

  const char *get_address() const { return address; }
  unsigned get_listen_port() const { return listen_port; }

  void send_mcast( MTPPacket &packet);
  void send_ucast_reply( MTPPacket &packet, u16 port);

private:
    map_type hmap;
};

//----------------------------------------------------------PacketMgrFactory--
//
class PacketMgrFactory {
  omni_mutex guard;
  static int ttl;
  static unsigned listen_port;
  static sockaddr send_addr;
  static unsigned send_addr_len;
  static unsigned address_range;
  unsigned addresses_in_use;
  
#ifdef __WIN32__
public:
#endif
  struct cmpmgr {
    int operator() ( const PacketEndpoint *f1, 
		     const PacketEndpoint *f2) const {
      if (f1->ipaddress() < f2->ipaddress()) return 0;
      else return 1;
    }
  };
public:
  static void init( int &argc, char ** &argv);

  PacketMgrFactory() : addresses_in_use(0) {}

  PacketMgr *find_or_create( const char *endpoint_id);
  PacketMgr *assign_channel();
private:
  XSTD::set<PacketEndpoint*,cmpmgr> mgrs;
};

extern PacketMgrFactory packetmgrfactory;

class MTPConnectionIdMgr {
  u32 last_id;
  omni_mutex guard;
public:
  MTPConnectionIdMgr() : last_id(0) {}
  
  u32 assign_id() {
    omni_mutex_lock oml(guard);
#ifdef __WIN32__
    return (last_id+=2)+((unsigned) _getpid() *4096);
#else
    return (last_id+=2)+((unsigned) getpid() *4096);
#endif
  }
};

class MTPError {};

extern MTPConnectionIdMgr mtpidmgr;


#endif 

// __MTP_HH__





