// $Id: mtpsource.hh,v 1.1 1998/11/11 05:13:16 mencda Exp mencda $
//
// $Source: /home/mencda/thesis/mtp/mtpsource.hh,v $
//
// $Log: mtpsource.hh,v $
// Revision 1.1  1998/11/11 05:13:16  mencda
// Initial revision
//
//

#ifndef __MTPSOURCE_HH__
#define __MTPSOURCE_HH__

#include "mtp.hh"
#include "mcast.hh"
#include <deque>

static const unsigned long OUT_BUFFER_LEN= 10240;

struct RetentionPacket {
  Time time;
  MTPPacket packet;

  RetentionPacket() {}
  RetentionPacket( MTPPacket &_packet) : packet( _packet) {}
};

class MTPSource : 
  private IncomingPacketHandler,
  public McastSource
{
  omni_mutex guard;
  omni_condition cond_trans, cond_full, cond_full_eom, 
    cond_sender, cond_wait, cond_init;

  omni_mutex message_guard;

  u32 my_ucast_id;
  MTPWeb web;
  TSAP master;
  MTPPacket resp_packet;

  typedef XSTD::deque<RetentionPacket> RetentionBuffer;
  RetentionBuffer retention_buffer;
  
  int first_req;
  int eom; // end of message
  int has_token;
  MTPPacket out_packet;
  unsigned char *out_buffer;
  unsigned long out_buffer_len;
  unsigned long out_buffer_pos;   
  unsigned long out_buffer_usage; 

  PacketMgr &packetmgr;

  enum { STATE_INIT, STATE_FAILED, STATE_ESTABLISHED } state;

  void flush();                                            // may block
  void build_packet( u8 data_modifier, unsigned long len);
  void reply_nak_deny( MTPPacket &packet);
  void handle_nak( MTPPacket &packet);
  void send_data_packet();
  void send_cancel_packet();
  void send_token_request();

  static void static_run( void *args) {
    ((MTPSource*)args)->my_run();
  }
  void my_run();
  void init();
  virtual void handle( MTPPacket &packet);
  void got_token( MTPPacket &packet);

public:
  MTPSource( PacketMgr &a_packetmgr) :
    McastSource( NULL),
    cond_trans( &guard), cond_full( &guard), cond_full_eom( &guard),
    cond_sender( &guard), cond_wait( &guard), cond_init( &guard),
    state( STATE_INIT),
    out_buffer( new unsigned char[OUT_BUFFER_LEN]),
    out_buffer_len( OUT_BUFFER_LEN),
    out_buffer_pos( 0),
    out_buffer_usage( 0),
    first_req( 1), has_token( 0),
    eom(0),
    packetmgr( a_packetmgr) {
    init();
  }

  ~MTPSource();
	 
  void run() {
    (new omni_thread( static_run, this))->start();
  }

  PacketMgr& get_packetmgr() const { return packetmgr; }

  void start_message() { message_guard.lock(); }
  void end_message() { flush(); message_guard.unlock(); }  // may block
  void message_data( void *data, unsigned long len);       // may block
  void write( void *data, unsigned long len) {
    message_data( data, len);
  } 
};


#endif 

// __MTPSOURCE_HH__




